//
//  GifMaker.m
//  GifMaker
//
//  Copyright 2006 T Endo. All rights reserved.
//

#include <time.h>
#include <stdlib.h>

// use GD library to output GIF images
#include <gd.h>

#import "GifMaker.h"
#import "GifObjects.h"

int convertToDecimal( char hex ) {
	if ( hex >= '0' && hex <= '9' ) {
		return (int)( hex - '0' );
	}
	if ( hex >= 'a' && hex <= 'z' ) {
		hex -= ( 'a' - 'A' );
	}
	if ( hex >= 'A' && hex <= 'F' ) {
		return 10 + ( hex - 'A' );
	}
	return 0;
}


@implementation GifMaker
-(id)init {
	[ super init ];
	self->width = 320;
	self->height = 240;
	self->delay = 10;
	self->smoothness = 4;
	self->vibration = 4;
	self->shadowOffset[ 0 ] = 4;
	self->shadowOffset[ 1 ] = 4;
	self->shadow[ 0 ] = 192;
	self->shadow[ 1 ] = 192;
	self->shadow[ 2 ] = 192;
	self->background[ 0 ] = 255;
	self->background[ 1 ] = 255;
	self->background[ 2 ] = 255;
	self->transparent = NO;
	self->processing = NO;
	self->framesPerLoop = 8;
	self->maxLineDivision = 10;
	srand( time( NULL ) );
	return self;
}
-(id)initWithWebView: ( WebView*) view {
	[ self init ];
	return self;	
}

-(void)dealloc {
	[ super dealloc ];
}

-(void)windowScriptObjectAvailable:(WebScriptObject*)ws {
//	NSLog( @"available" );
	[ ws setValue:self forKey:@"GifMaker" ];
}

+(NSString*)webScriptNameForSelector:(SEL)aSel {
//	NSLog( @"%@", aSel );
	if ( aSel == @selector(log:) ) {
		return @"log";
	} else if ( aSel == @selector(outputGif) ) {
		return @"outputGif";
	} else if ( aSel == @selector(setSmoothness:) ) {
		return @"setSmoothness";
	} else if ( aSel == @selector(setVibration:) ) {
		return @"setVibration";
	} else if ( aSel == @selector(setDelay:) ) {
		return @"setDelay";
	} else if ( aSel == @selector(setSize::) ) {
		return @"setSize";
	} else if ( aSel == @selector(setBackgroundColor:::) ) {
		return @"setBackgroundColor";
	} else if ( aSel == @selector(setShadowColor:::) ) {
		return @"setShadowColor";
	} else if ( aSel == @selector(setShadowOffset::) ) {
		return @"setShadowOffset";
	} else if ( aSel == @selector(outputWobblingGif:) ) {
		return @"outputWobblingGif";
	} else if ( aSel == @selector(removeFile:) ) {
		return @"removeFile";
	} else if ( aSel == @selector(setTransparent:) ) {
		return @"setTransparent";
	} else if ( aSel == @selector(showProcessing:) ) {
		return @"showProcessing";
	} else if ( aSel == @selector(setFramesPerLoop:) ) {
		return @"setFramesPerLoop";
	}
	return nil;
}

-(void)setFramesPerLoop: (int)frames {
	framesPerLoop = frames;
}

-(void)showProcessing:(BOOL) flag {
	processing = flag;
}


-(void)setShadowOffset: (int)x :(int)y {
	self->shadowOffset[ 0 ] = x;
	self->shadowOffset[ 1 ] = y;
}

-(void)setBackgroundColor: (int)r :(int)g :(int)b {
	background[ 0 ] = r;
	background[ 1 ] = g;
	background[ 2 ] = b;
}

-(void)setShadowColor: (int)r: (int)g :(int)b {
	shadow[ 0 ] = r;
	shadow[ 1 ] = g;
	shadow[ 2 ] = b;
}

-(void)setTransparent: (BOOL) flag {
	self->transparent = flag;
}

+(BOOL)isSelectorExcludedFromWebScript:(SEL)aSel {
	return NO;
}

+(BOOL)isKeyExcludedFromWebScript:(const char*)k {
	return YES;
}

-(void)log: (NSString*) message {
	fprintf( stderr, "%s\n", [ message UTF8String ] );
}

-(void)setSize: (int)aWidth :(int)aHeight {
	width = aWidth;
	height = aHeight;
}

-(void)setDelay: (int)count {
	delay = count;
}

-(void)setSmoothness: (int) sm {
	smoothness = sm;
}

-(void)setVibration: (int)vib {
	vibration = vib;
}

-(void)resolveColor : (NSString*)text rgb:(int*)values {
	int i;
	const char* str = [ text cString ];
	values[ 0 ] = 0;
	values[ 1 ] = 0;
	values[ 2 ] = 0;
	if ( *str == '#' ) {
		int length = [ text length ] - 1;
		if ( length <= 3 ) {
			for ( i = 0; i < length; i++ ) {
				values[ i ] = convertToDecimal( str[ i + 1 ] );
			}
		} else {
			for ( i = 0; i < length / 2; i++ ) {
				values[ i ] = convertToDecimal( str[ i * 2 + 1 ] ) * 16 + convertToDecimal( str[ i * 2 + 2 ] );
			}
		}
	}
}

-(NSArray*) convertToLines: (WebScriptObject*) lines {
	int i;
	int rgb[ 3 ];
	int count = [ [ lines valueForKey:@"length" ] intValue ];
	NSMutableArray* allLines = [ [ NSMutableArray alloc ] init ];
	for ( i = 0; i < count; i++ ) {
		WebScriptObject* line = [ lines valueForKey:[ NSString stringWithFormat:@"%d", i ] ];		
		// get properties
		id line_width = [ line valueForKey: @"width" ];
		id line_color = [ line valueForKey: @"hexcolor" ];
		id line_points = [ line valueForKey: @"data" ];		
		if ( line_width == nil || line_color == nil || line_points == nil ) continue;
		int thickness = [ line_width intValue ];
		[ self resolveColor:[ line_color description] rgb:rgb ];
		int num_points = [ [ line_points valueForKey: @"length" ] intValue ];
		int j;
		NSMutableArray* points = [ [ NSMutableArray alloc ] init ];
		j = 0;
		while ( j < num_points ) {
			id point = [ line_points valueForKey: [ NSString stringWithFormat :@"%d", j ] ];
			int xpos = [ [ point valueForKey: @"0" ] intValue ];
			int ypos = [ [ point valueForKey: @"1" ] intValue ];
			[ points addObject:[ LinePoint point:xpos :ypos ] ];
			j += self->smoothness;
		}
		// last point
		if ( j != num_points - 1 ) {
			j = num_points - 1;
			id point = [ line_points valueForKey: [ NSString stringWithFormat :@"%d", j ] ];
			int xpos = [ [ point valueForKey: @"0" ] intValue ];
			int ypos = [ [ point valueForKey: @"1" ] intValue ];
			[ points addObject:[ LinePoint point:xpos :ypos ] ];
		}
		[ allLines addObject:[ [ GifLine alloc ] initWithData:points width:thickness red:rgb[ 0 ] green:rgb[ 1 ] blue:rgb[ 2 ] ] ];
	}
	[ allLines retain ];
	return allLines;
}

-(void) drawGraphics :(gdImagePtr)image :(NSArray*)lines :(GifColorTable*)cTable {
	int i;
	int rcol, gcol, bcol;
	BOOL draw_shadow = YES;
	int bgcol = [ cTable getPalette:background[ 0 ]: background[ 1 ]: background[ 2 ] ];
	int shcol = [ cTable getPalette:shadow[ 0 ]: shadow[ 1 ]: shadow[ 2 ] ];
	int sx = shadowOffset[ 0 ];
	int sy = shadowOffset[ 1 ];
	if ( sx == 0 && sy == 0 ) {
		draw_shadow = NO;
	}
	int seed = rand();
	srand( seed );
	int vib_steps = self->vibration;
	
	// clear
	gdImageFilledRectangle( image, 0, 0, self->width, self->height, bgcol );
	
	// shadow
	if ( draw_shadow ) {
		for ( i = 0; i < [ lines count ]; i++ ) {
			GifLine* gl = [ lines objectAtIndex: i ];
			int num = [ gl getPointNumber ];
			if ( num < 2 ) continue;
			int x0 = [ gl getX: 0 ];
			int y0 = [ gl getY:0 ];
			int vx0 = rand() % vib_steps - self->vibration;
			int vy0 = rand() % vib_steps - self->vibration;
			int j;
			x0 += vx0;
			y0 += vy0;
			gdImageSetThickness( image, [ gl getThickness ] );		
			for ( j = 1; j < num; j++ ) {
				int x1 = [ gl getX: j ];
				int y1 = [ gl getY: j ];
				int vx = rand() % vib_steps - self->vibration;
				int vy = rand() % vib_steps - self->vibration;
				x1 += vx;
				y1 += vy;
				gdImageLine( image, x0 + sx, y0 + sy, x1 + sx, y1 + sy, shcol );
				x0 = x1;
				y0 = y1;
			}
		}	
	}
	
	srand( seed ); // use same seed number to set the shadow position aligned
	// lines
	for ( i = 0; i < [ lines count ]; i++ ) {
		GifLine* gl = [ lines objectAtIndex: i ];
		int num = [ gl getPointNumber ];
		if ( num < 2 ) continue;
		int x0 = [ gl getX: 0 ];
		int y0 = [ gl getY:0 ];
		int vx0 = rand() % vib_steps - self->vibration;
		int vy0 = rand() % vib_steps - self->vibration;
		int j;
		int thickness = [ gl getThickness ];
		int r = thickness / 2;
		[ gl getColor: &rcol :&gcol: &bcol ];
		int color = [ cTable getPalette:rcol: gcol :bcol ];
		gdImageSetThickness( image, thickness );
		BOOL round_corner = thickness >= 4;
		x0 += vx0;
		y0 += vy0;
		if ( round_corner ) {
			gdImageFilledEllipse( image, x0, y0, r, r, color );
		}
		for ( j = 1; j < num; j++ ) {
			int x1 = [ gl getX: j ];
			int y1 = [ gl getY: j ];
			int vx = rand() % vib_steps - self->vibration;
			int vy = rand() % vib_steps - self->vibration;
			x1 += vx;
			y1 += vy;
			gdImageLine( image, x0, y0, x1, y1, color );
			if ( round_corner ) {
				gdImageFilledEllipse( image, x1, y1, r, r, color );
			}
			x0 = x1;
			y0 = y1;
		}
	}	
	
}

-(NSString*)outputWobblingGif : (WebScriptObject*) lines {
	int i;
	NSArray* allLines = [ self convertToLines: lines ];	
	GifColorTable* color_table = [ [ GifColorTable alloc ] init ];
	
	/* create gif */
	gdImagePtr im;
	//gdImagePtr last, cim;
	FILE* out;
	//int i;
	NSString* filename = [ GifMaker createTemporaryFile:@".gif" ];
	NSLog( @"output filename: %@", filename );
	out = fopen( [ filename UTF8String ], "wb" );
	im = gdImageCreate( self->width, self->height );

	// allocate  color
	int bgcol = gdImageColorAllocate( im, self->background[ 0 ], self->background[ 1 ], self->background[ 2 ] );
	
	int shcol = gdImageColorAllocate( im, self->shadow[ 0 ], self->shadow[ 1 ], self->shadow[ 2 ] );
	[ color_table addPalette:background[ 0 ]: background[ 1 ]: background[ 2 ]: bgcol ];
	[ color_table addPalette:shadow[ 0 ]: shadow[ 1 ]: shadow[ 2 ]: shcol ];
	int rcol, gcol, bcol;
	if ( self->transparent ) {
		gdImageColorTransparent( im, bgcol );
	}
	for ( i = 0; i < [ allLines count ]; i++ ) {
		GifLine* gl = [ allLines objectAtIndex: i ];
		[ gl getColor: &rcol :&gcol: &bcol ];
		[ color_table addPalette:rcol: gcol: bcol: gdImageColorAllocate( im, rcol, gcol, bcol ) ];		
	}

	// clear 
	gdImageRectangle( im, 0, 0, self->width, self->height, bgcol );
	gdImagePtr image_prev = NULL;
	
	gdImageGifAnimBegin( im, out, 1, 0 );

	// show drawing process
	if ( self->processing ) {
		// drawing all lines
		for ( i = 0; i < [ allLines count ]; i++ ) {
			int j;
			NSMutableArray* lines_processing = [ NSMutableArray arrayWithArray: [ allLines subarrayWithRange:NSMakeRange( 0, i ) ] ];
			GifLine* line_drawing = (GifLine*)[ allLines objectAtIndex: i ];
			int num_points = [ line_drawing getPointNumber ];
			int steps = 1;
			int num_end = num_points;
			// skip points if there are too many points
			if ( num_points > self->maxLineDivision ) {
				steps = num_points / self->maxLineDivision;
			}
			if ( num_points % steps != 0 ) {
				num_end = num_points + steps;
			}

			// draw intermediate states
			for ( j = steps; j <= num_end; j+= steps ) {
				int num_end = j + 1;
				if ( num_end > num_points ) {
					num_end = num_points;
				}
				GifLine* line_intermediate = [ line_drawing subLine:0 : num_end ];
				[ lines_processing addObject: line_intermediate ];
				gdImagePtr image_cloned = gdImageCreate( self->width, self->height );
				gdImagePaletteCopy( image_cloned, im );
				if ( self->transparent ) {
					gdImageColorTransparent( image_cloned, bgcol );
				}
				[ self drawGraphics: image_cloned :lines_processing :color_table];
				gdImageGifAnimAdd( image_cloned, out, 0, 0, 0, self->delay, gdDisposalNone, image_prev );
				if ( image_prev ) {
					gdImageDestroy( image_prev );
				}
				image_prev = image_cloned;
				[ lines_processing removeObject: line_intermediate ];
			}			
		}
	}
	for ( i = 0; i < self->framesPerLoop; i++ ) {
		gdImagePtr image_cloned = gdImageCreate( self->width, self->height );
		gdImagePaletteCopy( image_cloned, im );
		if ( self->transparent ) {
			gdImageColorTransparent( image_cloned, bgcol );
		}
		[ self drawGraphics: image_cloned :allLines :color_table];
		gdImageGifAnimAdd( image_cloned, out, 0, 0, 0, self->delay, gdDisposalNone, image_prev );
		if ( image_prev ) {
			gdImageDestroy( image_prev );
		}
		image_prev = image_cloned;
	}

	if ( image_prev != NULL ) {
		gdImageDestroy( image_prev );
	}
	
	gdImageGifAnimEnd( out );
	fclose( out );
	gdImageDestroy( im );
	[ filename retain ];
	[ allLines release ];
	return filename;
}


-(void)removeFile: (NSString*) filename {
	NSFileManager* fm = [ NSFileManager defaultManager ];
	NSLog( @"removing %@", filename );
	if ( [ fm fileExistsAtPath: filename ] == NO ) {
		NSLog( @"%@ does not exist", filename );
	}
	if ( [ fm removeFileAtPath:filename handler:nil ] == NO ) {
		NSLog( @"%@ cannot be removed", filename );
	}
}

+(NSString*) createTemporaryFile : (NSString*) suffix {
  NSString* dir_temp = NSTemporaryDirectory();
  NSMutableString* filename = [ [ NSMutableString alloc ] init ];
  NSMutableString* path_temp;
  NSFileManager* fm = [ NSFileManager defaultManager ];
  BOOL is_dir = NO;
  const char* characters = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  int num_chars = strlen( characters );
  int i;

  for ( i = 0; i < 6; i++ ) {
    int num = rand() % num_chars;
    [ filename appendFormat: @"%c", characters[ num ] ];
  }

  for ( ;; ) {
    int num;
    BOOL exists = [ fm fileExistsAtPath: [ NSString stringWithFormat: @"%@/%@%@", dir_temp, filename, suffix ] isDirectory: &is_dir ];
    if ( exists == NO && is_dir == NO ) {
      break;
    } 
    num = rand() % num_chars;
    [ filename appendFormat: @"%c", characters[ num ] ];
    is_dir = NO;
  }

  path_temp = [ NSString stringWithFormat: @"%@/%@%@", dir_temp, filename, suffix ];
  [ path_temp retain ];
  return path_temp;
}

@end
